# EventScripts Minecraft Example
# Copyright 2011, Mattie Casper
#
# magicblocks.py
#   Replace /warp, /sethome, and /home with some new commands
#  

import esm
import esm.blocklib
reload(esm.blocklib)
import os
import utils
import time

jumpcounters = {}
class PlayerJumpTracker():
    def __init__(self, player):
        self.player = player
        self.lastmove = 0
        self.jumpcount = 0
        self.lastblock = player.location
    def move(self, mfrom, mto):
        id = esm.blockat(mto.blockx, mto.blocky, mto.blockz, mto.getWorld()).typeid
        if id in (esm.items.water, esm.items.ladder):
            # not jumping if there's a ladder or water involved
            self.reset()
        else:
            #esm.output("%f,%f,%f -> %f,%f,%f" %(mfrom.x, mfrom.y, mfrom.z, mto.x, mto.y, mto.z))
            #if self.lastblock.blockx-mto.blockx !=0 or self.lastblock.blocky-mto.blocky != 0 or self.lastblock.blockz-mto.blockz != 0:
            #    esm.output("%d,%d,%d" % (self.lastblock.blockx-mto.blockx, self.lastblock.blocky-mto.blocky, self.lastblock.blockz-mto.blockz))
            if self.lastblock.blockx==mto.blockx and self.lastblock.blockz==mto.blockz and self.lastblock.blocky != mto.blocky:
                if self.lastblock.blocky == mto.blocky - 1:
                    self._jumpup()
                elif self.lastblock.blocky - 1 == mto.blocky:
                    self._falldown()
                else:
                    self.reset()
            elif self.lastblock.blockx==mto.blockx and self.lastblock.blockz==mto.blockz and self.lastblock.blocky == mto.blocky:
                pass
            else:
                self.reset()
        self.lastblock = mto
        return self.jumpcount
    def _jumpup(self):
        if self.lastmove == -1:
            self.jumpcount += 1
        self.lastmove = 1
    def _falldown(self):
        self.lastmove = -1
    def reset(self):
        self.lastmove = 0
        self.jumpcount = 0



from java import io
class MagicBlocks():
    def __init__(self, cfg):
        self.jumpcounters = {}
        self.warpcooldown = {}
        self.homesdat = cfg.getString("magicblocks-homes-datafile", "homes.dat")
        self.warpsdat = cfg.getString("magicblocks-hopstop-datafile", "warps.dat")
        self.homewardbound_enabled = cfg.getBoolean("magicblocks-homewardbound-enabled", True)
        self.homebound_jumpcost = cfg.getInt("magicblocks-homewardbound-usage-jumpcost", 7)
        self.homerun_enabled = cfg.getBoolean("magicblocks-homerun-enabled", True)
        self.homerun_cost = cfg.getInt("magicblocks-homerun-usage-redstonecost", 4)
        self.hopstop_enabled = cfg.getBoolean("magicblocks-hopstop-enabled", True)
        self.hopstop_jumpcost = cfg.getInt("magicblocks-hopstop-usage-jumpcost", 2)
        self.hopstop_createcost = cfg.getInt("magicblocks-hopstop-create-redstonecost", 50)
        self.hopstop_hopcost = cfg.getInt("magicblocks-hopstop-usage-redstonecost-per-segment", 1)
        self.hopstop_world_hopcost = cfg.getInt("magicblocks-hopstop-usage-redstonecost-across-worlds", 30)
        self.hopstop_hoplength = cfg.getInt("magicblocks-hopstop-segment-distance", 100)
        if os.path.exists(self.homesdat):
            self.homes = self.loadDict(self.homesdat)
        else:
            self.homes = {}
        if os.path.exists(self.warpsdat):
            self.warps = self.loadDict(self.warpsdat)
        else:
            self.warps = {}
    def saveDict(self, x,fname):
        outs=io.ObjectOutputStream(io.FileOutputStream(fname))
        outs.writeObject(x)
        outs.close()
    def loadDict(self, fname):
        ins=io.ObjectInputStream(io.FileInputStream(fname))
        x=ins.readObject()
        ins.close()
        return x
    def trackPlayer(self, player):
        name = player.getName()
        self.jumpcounters[name] = PlayerJumpTracker(player)
        return True
    def warpHome(self, player):
        name = player.getName()
        loc = (0,0,0)
        if name not in self.homes:
            sloc = esm.world.spawnlocation
            loc = (sloc.x, sloc.y-1, sloc.z)
        else:
            loc = self.homes[name]
        if len(loc) > 3:
            wo = esm.server.getWorld(loc[3])
        else:
            wo = esm.world
        location = esm.location(loc[0]+0.5, loc[1]+1, loc[2]+0.5, player.location.yaw, player.location.pitch, wo)
        player.teleportTo(location)

    def useBasequick(self, player):
        if self.homerun_enabled and utils.spendRedstone(player, self.homerun_cost):
            self.warpHome(player)
        return True

    def useBasejump(self, player, lfrom, lto):
        if not self.homewardbound_enabled and not self.hopstop_enabled:
            return True
        name = player.getName()
        if name not in self.jumpcounters:
            self.trackPlayer(player)
        jumped = self.jumpcounters[name].move(lfrom, lto)
        if self.homewardbound_enabled and player.iteminhand.typeid == esm.items.air:
            if jumped >= self.homebound_jumpcost-1:
                self.jumpcounters[name].reset()
                self.warpHome(player)
                return False
        if self.hopstop_enabled and jumped >= self.hopstop_jumpcost-1:
            if self.useWarpRoom(player, lto):
                self.jumpcounters[name].reset()
                return False
        return True
    
    def useBasecamp(self, player, block):
        if not self.homewardbound_enabled and not self.homerun_enabled:
            return True
        name = player.getName()
        if name in self.homes:
            # destroy home
            oldhome = self.homes[name]
            if len(oldhome) > 3:
                world = esm.server.getWorld(oldhome[3])
            else:
                world = esm.world
            world.loadChunk(block.x, block.z)
            bl = esm.blockat(int(oldhome[0]), int(oldhome[1]), int(oldhome[2]), world)
            bl.typeid = esm.items.cobblestone
        # surely they were still holding the compass
        inv = player.inventory
        inv.clear(inv.helditemslot)
        # set block to obsidian
        block.typeid = esm.items.obsidian
        self.homes[name] = (block.x, block.y, block.z, str(block.world.name))
        self.saveDict(self.homes, self.homesdat)
        # save block+1 as player home spot
        return False

    def checkWarpStructure(self, block, style=None):
        # checkWarpStructure assumes it gets the core block
        wo, ox, oy, oz = block.getWorld(), block.getX(), block.getY(), block.getZ()
        for x in range(ox-1, ox+2):
            for z in range(oz-1, oz+2):
                wo.loadChunk(x, z)
                #esm.world.getChunkAt(x, z)
                if x == ox and z == oz:
                    if style is None:
                        continue
                    else:
                        # testing a working structure
                        if self.getWarpType(esm.blockat(x, oy, z, wo)) != style:
                            esm.output('failed1')
                            return 0
                        if esm.blockidat(x, oy+1, z, wo) != esm.items.redstonewire:
                            esm.output('failed_redstone')
                            esm.output(esm.blockidat(x, oy+1, z, wo))
                            return 0
                    continue
                if esm.blockidat(x, oy, z, wo) != esm.items.obsidian:
                    esm.output("not obsid %d, %d, %d -- was %d" %(x,oy,z, esm.blockidat(x, oy, z, wo)))
                    return 0
        return 1

    def removeWarpPoint(self, block, typ):
        wo = block.getWorld()
        wo.loadChunk(block.getX(), block.getZ())
        # removeWarpPoint assumes it gets the core block
        esm.output("MagicItems[HopStop]: Removing warp point" + str(typ) + " " + str((block.getX(), block.getY(), block.getZ(), wo.getName())))
        pt = (block.getX(), block.getY(), block.getZ(), wo.getName())
        if pt in self.warps[typ]:
            self.warps[typ].remove(pt)
        else:
            pt = (block.getX(), block.getY(), block.getZ())
            self.warps[typ].remove(pt)
        if len(self.warps[typ]) < 1:
            del self.warps[typ]
        self.saveDict(self.warps, self.warpsdat)

    def getContentsWarpType(self, inv):
        typ = []
        try:
            if inv.getSize() > 0:
                contents = inv.getContents()
                i = 0
                for item in contents:
                    if item.getTypeId() != 0:
                        data = item.getData()
                        data = data.getData() if data else None
                        typ.append((i, item.getTypeId(), data, item.getAmount(),))
                    i+=1
        except:
            esm.output("[MagicItems] Warning: Couldn't read warp target contents.")
        return tuple(typ)

    def getWarpType(self, block):
        id = block.getTypeId()
        data = block.getData()
        if id == esm.items.note_block:
            st = block.getState()
            typ = (id, data, st.getNote())
            st.play()
        elif id in [esm.items.chest, esm.items.dispenser, esm.items.furnace]:
            st = block.getState()
            items = self.getContentsWarpType(st.getInventory())
            if len(items) < 1:
                typ = id
            elif id == esm.items.furnace:
                typ = (id, items)
            else:
                typ = (id, data, items)
        elif data and data != 0:
            typ = (id, data)
        else:
            typ = id
        return typ

    def createWarpPoint(self, block):
        typ = self.getWarpType(block)
        esm.output(typ)
        wo = block.getWorld()
        pt = (block.getX(), block.getY(), block.getZ(), wo.getName())
        if typ and typ in self.warps:
            if pt and pt not in self.warps[typ]:
                esm.output("MagicItems[HopStop]: Appending" + str(pt) +" to " + str(typ) + " in " + str(self.warps[typ]))
                self.warps[typ].append(pt)
                esm.blockat(block.getX(), block.getY()+1, block.getZ(), wo).typeid = esm.items.redstonewire
        elif typ:
            esm.output("MagicItems[HopStop]: Adding " + str(pt) +" to " + str(typ))
            self.warps[typ] = [pt]
            esm.blockat(block.getX(), block.getY()+1, block.getZ(), wo).typeid = esm.items.redstonewire
        self.saveDict(self.warps, self.warpsdat)

    # called on destroy block
    def useWarpBind(self, player, block):
        esm.output(str(player.getItemInHand())+"."+ str(esm.items.redstonedust))
        if self.hopstop_enabled and player.iteminhand.typeid == esm.items.redstonedust:
            if self.warpcooldown.get(player.getName(), 0) < time.time() - 2 and self.checkWarpStructure(block)==1 and utils.spendRedstone(player, self.hopstop_createcost):
                self.createWarpPoint(block)
                self.warpcooldown[player.getName()] = time.time()
                return False
            else:
                return True
        return True

    def _warpActivate(self, player, bl2):
        if self.checkWarpStructure(bl2) != 1:
            return True
        warptype = self.getWarpType(bl2)
        #esm.output(warptype)
        #esm.output(self.warps.keys())
        if warptype in self.warps and self.warpcooldown.get(player.getName(), 0) < time.time() - 2:
            esm.output("warptype found")
            wo = bl2.getWorld()
            spotnew = (bl2.getX(), bl2.getY(), bl2.getZ(), wo.getName())
            spotold = (bl2.getX(), bl2.getY(), bl2.getZ())
            #esm.output("all warps of the type: " + str(self.warps[warptype]))
            #esm.output("where we are now " + str(spot))
            spot = None
            if spotnew in self.warps[warptype]:
                spot = spotnew
            elif wo.getName() == esm.world.getName() and spotold in self.warps[warptype]:
                spot = spotold
            if spot is not None:
                loc = spot
                while len(self.warps[warptype]) > 1:
                    ind = self.warps[warptype].index(loc)
                    try:
                        loc = self.warps[warptype][ind+1]
                    except:
                        loc = self.warps[warptype][0]
                    if len(loc) < 4:
                        bl3 = esm.blockat(loc[0], loc[1], loc[2])
                    else:
                        wo3 = esm.server.getWorld(loc[3])
                        bl3 = esm.blockat(loc[0], loc[1], loc[2], wo3)
                    ret = self.checkWarpStructure(bl3, warptype)
                    if ret == -1:
                        # checking it will load the chunks we need
                        esm.output("MagicItems[HopStop]: Queueing because chunk wasn't loaded, let's try again later.")
                        esm.queue(self._warpActivate, (player, bl2))
                    if ret == 0:
                        self.removeWarpPoint(bl3, warptype)
                        loc = spot
                    else:
                        break
                esm.output("MagicItems[HopStop]: Trying to teleport %s to %s" %(player.getName(), str(loc)))
                worldname1 = spot[3] if len(spot) > 3 else esm.world.getName()
                worldname2 = loc[3] if len(loc) > 3 else esm.world.getName()
                if worldname1 != worldname2:
                    cost = self.hopstop_world_hopcost
                else:
                    cost = int(esm.blocklib.distance(spot, loc)/self.hopstop_hoplength)*self.hopstop_hopcost
                if cost<=0 or utils.spendRedstone(player, cost):
                    wo = esm.server.getWorld(worldname2)
                    location = esm.location(loc[0]+0.5, loc[1]+1, loc[2]+0.5, player.location.yaw, player.location.pitch, wo)
                    player.teleportTo(location)
                else:
                    health = player.getHealth()
                    if health <= 1:
                        player.setHealth(health+1)
                    player.setHealth(player.getHealth()-1)
                self.warpcooldown[player.getName()] = time.time()
                return True
        return False

    # called on player move
    def useWarpRoom(self, player, lto):
        bl = esm.blockat(int(lto.blockx), int(lto.blocky), int(lto.blockz), lto.getWorld())
        if bl.typeid == esm.items.redstonewire:
            bl2 = esm.blockat(bl.x, bl.y-1, bl.z, bl.getWorld())
            return self._warpActivate(player, bl2)
        return False
