# EventScripts Minecraft Example
# Copyright 2010, Mattie Casper
#
# magicpick.py
#   Helps you find rare ore in the area
#
import esm
import esm.blocklib as blocklib
import time
from esm import items
import magicitems
from threading import Timer

scanitems= [items.goldore, items.diamondore, items.redstoneore, items.ironore]

class MagicPick():
    def __init__(self, cfg):
        self.cooldowns = {}
        self.protectedblocks = {}
        pre = "magicpick-"
        # defaults
        self.radius = cfg.getInt(pre+"allorenothing-radius", 3)
        self.cooldown = cfg.getInt(pre+"allorenothing-cooldown", 10)
        self.scancost = cfg.getInt(pre+"allorenothing-scancost", 3)
    def _performResourceScan(self, block):
        # be prepared to revert the block
        x, y, z = block.getX(), block.getY(), block.getZ()
        origstat = block.getStatus()
        origtype = block.getType()
        # find something other than the ore hit
        finditems = list(scanitems)
        if origtype in finditems:
            finditems.remove(origtype)
        elif origtype == items.redstoneorealt:
            finditems.remove(items.redstoneore)
        # find one of the resource types
        find = blocklib.findFirstMatchingBlock(block, self.radius, finditems)
        findtype = items.bedrock
        if find is not None:
            findtype = find.getType()
        block.setType(findtype)
        block.update()
        self.addProtectedBlock(block)
        # schedule the block to revert
        def revertBlock(typ, stat, bx, by, bz):
            rbl = esm.server.getBlockAt(bx, by, bz)
            rbl.setStatus(stat)
            rbl.setType(typ)
            rbl.update()
            self.removeProtectedBlock(bx, by, bz)
        Timer(1.00, esm.queue, (revertBlock, (origtype, origstat, x, y, z))).start()

    def checkProtectedBlock(self, block):
        if (block.getX(), block.getY(), block.getZ()) in self.protectedblocks:
            return True
        return False

    def addProtectedBlock(self, block):
        x, y, z = block.getX(), block.getY(), block.getZ()
        if (x,y,z) in self.protectedblocks:
            self.protectedblocks[(x,y,z)] = self.protectedblocks[(x,y,z)] + 1
        else:
            self.protectedblocks[(x,y,z)] = 1

    def removeProtectedBlock(self, x, y, z):
        if (x,y,z) in self.protectedblocks:
            val = self.protectedblocks[(x,y,z)]
            self.protectedblocks[(x,y,z)] = val - 1
            if self.protectedblocks[(x,y,z)] <= 0:
                del self.protectedblocks[(x,y,z)]
        

    def useResourceScan(self, player, block):
        name = player.getName()
        if name in self.cooldowns:
            if self.cooldowns[name] > time.time() - self.cooldown:
                return True
        if magicitems.spendRedstone(player, self.scancost):
            self.cooldowns[name] = time.time()
            self._performResourceScan(block)
            return False
        return True

